"""
tests for the downloader file
"""
import os
# import sys
import unittest
import time
import shutil
import platform

# import custom python packages
# sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from wahoomc.downloader import older_than_x_days
from wahoomc.downloader import download_file
from wahoomc.downloader import build_osm_pbf_filepath
from wahoomc.downloader import download_tooling
from wahoomc.downloader import Downloader
from wahoomc import constants
from wahoomc.constants_functions import get_tooling_win_path
from wahoomc.geofabrik_json import GeofabrikJson


class TestDownloader(unittest.TestCase):
    """
    tests for the downloader file
    """

    def setUp(self):
        self.max_days_old = 14

        self.o_downloader = Downloader(24, False)

    def test_outdated_timestamp(self):
        """
        Test a outdated timestamp against 14 days ago
        """

        past = time.time() - (60 * 60 * 24 * 20)

        result = older_than_x_days(past, self.max_days_old)
        self.assertTrue(result)

    def test_outdated_timestamp_2(self):
        """
        Test a outdated timestamp against 14 days ago
        """

        result = older_than_x_days(1601097377.792748, self.max_days_old)
        self.assertTrue(result)

    def test_today_timestamp(self):
        """
        Test today timestamp against 14 days ago
        """

        today = time.time()

        result = older_than_x_days(today, self.max_days_old)
        self.assertFalse(result)

    def test_future_timestamp(self):
        """
        Test a timestamp in the future against 14 days ago
        """

        future = time.time() + (60 * 60 * 24 * 20)

        result = older_than_x_days(future, self.max_days_old)
        self.assertFalse(result)

    def test_download_geofabrik_file(self):
        """
        Test the download of geofabrik file via URL
        """
        path = os.path.join(constants.USER_DL_DIR, 'geofabrik.json')
        download_file(
            path, 'https://download.geofabrik.de/index-v1.json', False)

        self.assertTrue(os.path.exists(path))

    def test_download_geofabrik_file_2(self):
        """
        Test the download of geofabrik file via URL.
        check & download via built functions of downloader class
        """
        path = os.path.join(constants.USER_DL_DIR, 'geofabrik.json')

        if os.path.exists(path):
            os.remove(path)

        self.assertTrue(
            self.o_downloader.should_geofabrik_file_be_downloaded())

        self.o_downloader.download_geofabrik_file()

        self.assertTrue(os.path.exists(path))
        self.assertFalse(
            self.o_downloader.should_geofabrik_file_be_downloaded())

    def test_building_osm_pbf_filepaths(self):
        """
        Test if composing map file path on the device is correct
        "/" to "_" is done for "us/" countries
        """
        self.check_exp_agains_composed_map_file_path(
            'us/nebraska', 'us_nebraska')

        self.check_exp_agains_composed_map_file_path(
            'us/georgia', 'us_georgia')

        self.check_exp_agains_composed_map_file_path(
            'malta', 'malta')

        self.check_exp_agains_composed_map_file_path(
            'asia', 'asia')

        self.check_exp_agains_composed_map_file_path(
            'georgia', 'georgia')

    def test_download_osm_pbf_file(self):
        """
        Test the download of geofabrik file via URL
        """
        self.check_download_osm_pbf_file('malta', os.path.join(constants.USER_DL_DIR, 'maps',
                                                               'malta' + '-latest.osm.pbf'))

        self.check_download_osm_pbf_file('us/hawaii', os.path.join(constants.USER_DL_DIR, 'maps',
                                                                   'us_hawaii' + '-latest.osm.pbf'))

        self.check_download_osm_pbf_file('solomon-islands', os.path.join(constants.USER_DL_DIR, 'maps',
                                                                         'solomon-islands' + '-latest.osm.pbf'))

    def test_delete_not_existing_file(self):
        """
        Test if the removal of a not existing file raises a exception
        """

        path = os.path.join(constants.USER_DL_DIR, 'maps',
                            'malta' + '-latest.osm.pbf')

        if os.path.exists(path):
            os.remove(path)

        with self.assertRaises(FileNotFoundError):
            os.remove(path)

    def test_check_dl_needed_geofabrik(self):
        """
        Test if geofabrik file needs to be downloaded
        """
        path = os.path.join(constants.USER_DL_DIR, 'geofabrik.json')

        if os.path.exists(path):
            os.remove(path)

        self.o_downloader.max_days_old = 24
        # self.o_downloader.check_geofabrik_file()

        self.assertTrue(
            self.o_downloader.should_geofabrik_file_be_downloaded())

    def test_download_windows_files(self):
        """
        Test if Windows tooling files download is successful
        """
        if platform.system() == "Windows":
            path = constants.USER_TOOLING_WIN_DIR

            if os.path.exists(path):
                shutil.rmtree(path)

            self.assertFalse(os.path.exists(path))

            download_tooling()

            self.assertTrue(os.path.exists(path))

            self.assertTrue(
                os.path.exists(get_tooling_win_path('osmfilter.exe', in_user_dir=True)))

    def test_download_macos_files(self):
        """
        Test if macOS tooling files download is successful
        """
        if platform.system() != "Windows":
            path = os.path.join(str(constants.USER_DIR), '.openstreetmap', 'osmosis',
                                'plugins', 'mapsforge-map-writer-0.18.0-jar-with-dependencies.jar')

            if os.path.exists(path):
                os.remove(path)

            self.assertFalse(os.path.exists(path))

            download_tooling()

            self.assertTrue(os.path.exists(path))

    def check_exp_agains_composed_map_file_path(self, country, country_map_file_path):
        """
        helper function to check a given to-be-path against the composed on based on country
        """
        exp_path = os.path.join(constants.USER_DL_DIR, 'maps',
                                f'{country_map_file_path}' + '-latest.osm.pbf')

        composed_path = build_osm_pbf_filepath(country)

        self.assertEqual(exp_path, composed_path)

    def check_download_osm_pbf_file(self, country, path):
        """
        Test if the download of a input country created the wanted map file
        """
        o_geofabrik_json = GeofabrikJson()

        if os.path.exists(path):
            os.remove(path)

        map_file_path = build_osm_pbf_filepath(country)
        url = o_geofabrik_json.get_geofabrik_url(country)
        download_file(map_file_path, url, False)

        self.assertTrue(os.path.exists(path))


if __name__ == '__main__':
    unittest.main()



"""
functions and object for managing OSM maps
"""
#!/usr/bin/python
# pylint: skip-file

# import official python packages
import sys
import math
import logging
from shapely.geometry import Polygon, shape

# import custom python packages
from wahoomc.constants import special_regions, block_download
from wahoomc.geofabrik_json import GeofabrikJson

log = logging.getLogger('main-logger')


class InformalGeofabrikInterface:
    """Informal class for Geofabrik processing"""
    wanted_maps = []
    tiles = []
    border_countries = {}
    output = {}

    o_geofabrik_json = GeofabrikJson()

    def get_tiles_of_wanted_map_single(self, wanted_map) -> str:
        """Get the relevant tiles for ONE wanted country or X/Y coordinate"""
        pass

    def get_tiles_of_wanted_map(self) -> list:
        """
        calculates tiles of all input wanted-maps
        :returns: list of tiles
        """

        tiles_of_input = []

        for country in self.wanted_maps:
            tiles_of_input.extend(self.get_tiles_of_wanted_map_single(country))

        return tiles_of_input

    def find_needed_countries(self, bbox_tiles, wanted_map, wanted_region_polygon) -> dict:
        """find needed countries for requested country or X/Y combination"""
        pass

    def compose_bouding_box(self, bounds) -> dict:
        """calculate bounding box based on geometry or X/Y combination"""
        pass


class CountryGeofabrik(InformalGeofabrikInterface):
    """Geofabrik processing for countries"""

    def __init__(self, input_countries):
        """
        :param input: string with countries: 'malta' or 'malta, switzerland'
        :returns: object for country geofabrik processing
        """
        self.wanted_maps = []

        # input parameters
        input_countries = get_countries_from_input(input_countries)

        for country in input_countries:
            self.wanted_maps.append(self.o_geofabrik_json.translate_id_no_to_geofabrik(
                country))

    def get_tiles_of_wanted_map_single(self, wanted_map):
        """Overrides InformalGeofabrikInterface.get_tiles_of_wanted_map_single()"""
        # Check if wanted_map is in the json file and if so get the polygon (shape)
        wanted_map_geom = self.o_geofabrik_json.get_geofabrik_geometry(
            wanted_map)

        # convert to shape (multipolygon)
        wanted_region = shape(wanted_map_geom)

        # calc bounding box - the whole area to be created
        bbox = self.compose_bouding_box(wanted_region.bounds)

        # Build bounding box list of tiles - several X/Y combinations making up the area
        bbox_tiles = calc_bounding_box_tiles(bbox)

        # get all infos of these bounding box tiles
        tiles_of_input = self.find_needed_countries(
            bbox_tiles, wanted_map, wanted_region)

        return tiles_of_input

    def find_needed_countries(self, bbox_tiles, wanted_map, wanted_region_polygon) -> list:
        """Overrides InformalGeofabrikInterface.find_needed_countries()"""
        output = []

        geofabrik_regions = self.o_geofabrik_json.geofabrik_regions

        # itterate through tiles and find Geofabrik regions that are in the tiles
        counter = 1
        for tile in bbox_tiles:
            # Do progress indicator every 50 tiles
            if counter % 50 == 0:
                log.info(
                    'Processing tile %s of %s', counter, len(bbox_tiles)+1)
            counter += 1

            parent_added = 0
            force_added = 0

            # example contents of tile: {'index': 0, 'x': 130, 'y': 84, 'tile_left': 2.8125,
            # 'tile_top': 52.48278022207821, 'tile_right': 4.21875, 'tile_bottom': 51.6180165487737}
            # convert tile x/y to tile polygon lon/lat
            poly = Polygon([(tile["tile_left"], tile["tile_top"]), (tile["tile_right"],
                                                                    tile["tile_top"]), (tile["tile_right"], tile["tile_bottom"]),
                            (tile["tile_left"], tile["tile_bottom"]), (tile["tile_left"],
                                                                       tile["tile_top"])])

            # (re)initialize list of needed maps and their url's
            must_download_maps = []
            must_download_urls = []

            # itterate through countries/regions in the geofabrik json file
            for region, value in self.o_geofabrik_json.geofabrik_overview.items():
                regionname = region
                try:
                    parent = value['parent']
                except KeyError:
                    parent = ''
                rurl = value['pbf_url']
                rshape = shape(value['geometry'])

                # for debugging
                # if tile["x"] == 137 and (region == "denmark" or region == "germany" or region == "sweden"):
                #     print("stop for debugging")
                # if not xy_mode:

                # print (f'Processing region: {regionname}')

                # check if the region we are processing is needed for the tile we are processing

                # If currently processing country/region IS the desired country/region
                if regionname == wanted_map:
                    # Check if it is part of the tile we are processing
                    if rshape.intersects(poly):  # if so
                        # catch special_regions like (former) colonies where the map of the region is not fysically in the map of the parent country.
                        # example Guadeloupe, it's parent country is France but Guadeloupe is not located within the region covered by the map of France
                        if wanted_map not in special_regions:
                            # If we are proseccing a sub-region add the parent of this sub-region
                            # to the must download list.
                            # This to prevent downloading several small regions AND it's containing region
                            # we are processing a sub-regiongo find the parent region:
                            if parent not in geofabrik_regions and regionname not in geofabrik_regions:
                                # we are processing a sub-regiongo find the parent region
                                x_value = 0
                                # handle sub-sub-regions like unterfranken->bayern->germany
                                while parent not in geofabrik_regions:
                                    parent, child = self.o_geofabrik_json.get_geofabrik_parent_country(
                                        parent)
                                    if parent in geofabrik_regions:
                                        parent = child
                                        break
                                    if x_value > 10:  # prevent endless loop
                                        log.error(
                                            'Can not find parent map of region: %s', regionname)
                                        sys.exit()
                                    x_value += 1
                                if parent not in must_download_maps:
                                    must_download_maps.append(parent)
                                    must_download_urls.append(
                                        self.o_geofabrik_json.get_geofabrik_url(parent))
                                    # parent_added = 1
                            else:
                                if regionname not in must_download_maps:
                                    must_download_maps.append(regionname)
                                    must_download_urls.append(rurl)
                        else:
                            # wanted_map is a special region like Guadeloupe, France
                            if regionname not in must_download_maps:
                                must_download_maps.append(regionname)
                                must_download_urls.append(rurl)
                        # if there is an intersect, force the tile to be put in the output
                        force_added = 1
                    else:  # currently processing tile does not contain, a part of, the desired region
                        continue

                # currently processing country/region is NOT the desired country/region but might be
                # in the tile (neighbouring country)
                if regionname != wanted_map:
                    # check if we are processing a country or a sub-region.
                    # For countries only process other countries. also block special geofabrik sub regions
                    if parent in geofabrik_regions and regionname not in block_download:
                        # processing a country and no special sub-region
                        # check if rshape is subset of desired region. If so discard it
                        if wanted_region_polygon.contains(rshape):
                            # print (f'\t{regionname} is a subset of {wanted_map}, discard it')
                            continue
                        # check if rshape is a superset of desired region. if so discard it
                        if rshape.contains(wanted_region_polygon):
                            # print (f'\t{regionname} is a superset of {wanted_map}, discard it')
                            # if regionname not in must_download_maps:
                            #    must_download_maps.append (regionname)
                            #    must_download_urls.append (rurl)
                            #    parent_added = 1
                            continue
                        # Check if rshape is a part of the tile
                        if rshape.intersects(poly):
                            # print(f'\tintersecting tile: {regionname} tile={tile}')
                            if regionname not in must_download_maps:
                                must_download_maps.append(regionname)
                                must_download_urls.append(rurl)

            # If this tile contains the desired region, add it to the output
            # print (f'map= {wanted_map}\tmust_download= {must_download_maps}\tparent_added= {parent_added}\tforce_added= {force_added}')
            if wanted_map in must_download_maps or parent_added == 1 or force_added == 1:
                # first replace any forward slashes with underscores (us/texas to us_texas)
                must_download_maps = [sub.replace(
                    '/', '_') for sub in must_download_maps]
                output.append({'x': tile['x'], 'y': tile['y'], 'left': tile['tile_left'], 'top': tile['tile_top'],
                               'right': tile['tile_right'], 'bottom': tile['tile_bottom'], 'countries': must_download_maps, 'urls': must_download_urls})

        return output

    def compose_bouding_box(self, bounds):
        """Overrides InformalGeofabrikInterface.calc_bouding_box()"""
        # get bounding box
        (bbox_left, bbox_bottom, bbox_right, bbox_top) = bounds

        # convert bounding box to list of tiles at zoom level 8
        (top_x, top_y) = deg2num(bbox_top, bbox_left)
        (bot_x, bot_y) = deg2num(bbox_bottom, bbox_right)

        # and stay within the allowed tilenumber range!
        if top_x < 0:
            top_x = 0
        if top_x > 255:
            top_x = 255
        if top_y < 0:
            top_y = 0
        if top_y > 255:
            top_y = 255
        if bot_x < 0:
            bot_x = 0
        if bot_x > 255:
            bot_x = 255
        if bot_y < 0:
            bot_y = 0
        if bot_y > 255:
            bot_y = 255

        return {'top_x': top_x, 'top_y': top_y, 'bot_x': bot_x, 'bot_y': bot_y}


class XYGeofabrik(InformalGeofabrikInterface):
    """Geofabrik processing for X/Y coordinates"""

    def __init__(self, input_xy_coordinates):
        """
        :param input: string with xy-coordinates: 133/88 or 133/88,134/88
        :returns: object for xy geofabrik processing
        """
        # input parameters

        # use Geofabrik-URL to get the relevant tiles
        self.wanted_maps = get_xy_coordinates_from_input(input_xy_coordinates)

    def get_tiles_of_wanted_map_single(self, wanted_map):
        """Overrides InformalGeofabrikInterface.get_tiles_of_wanted_map_single()"""
        # calc bounding box - the whole area to be created
        bbox = self.compose_bouding_box(wanted_map)

        # Build bounding box list of tiles - several X/Y combinations making up the area
        bbox_tiles = calc_bounding_box_tiles(bbox)

        # convert X/Y combination to shape (multipolygon)
        wanted_region = self.compose_shape(bbox_tiles)

        # get all infos of these bounding box tiles
        tiles_of_input = self.find_needed_countries(
            bbox_tiles, wanted_map, wanted_region)

        return tiles_of_input

    def find_needed_countries(self, bbox_tiles, wanted_map, wanted_region_polygon) -> dict:
        """Overrides InformalGeofabrikInterface.find_needed_countries()"""
        output = []

        geofabrik_regions = self.o_geofabrik_json.geofabrik_regions

        log.info('Searching for needed maps, this can take a while.')

        # itterate through tiles and find Geofabrik regions that are in the tiles
        counter = 1
        for tile in bbox_tiles:
            # Do progress indicator every 50 tiles
            if counter % 50 == 0:
                log.info(
                    'Processing tile %s of %s', counter, len(bbox_tiles)+1)
            counter += 1

            parent_added = 0
            force_added = 0

            # example contents of tile: {'index': 0, 'x': 130, 'y': 84, 'tile_left': 2.8125,
            # 'tile_top': 52.48278022207821, 'tile_right': 4.21875, 'tile_bottom': 51.6180165487737}
            # convert tile x/y to tile polygon lon/lat
            poly = Polygon([(tile["tile_left"], tile["tile_top"]), (tile["tile_right"],
                                                                    tile["tile_top"]), (tile["tile_right"], tile["tile_bottom"]),
                            (tile["tile_left"], tile["tile_bottom"]), (tile["tile_left"],
                                                                       tile["tile_top"])])

            # (re)initialize list of needed maps and their url's
            must_download_maps = []
            must_download_urls = []

            # itterate through countries/regions in the geofabrik json file
            for region, value in self.o_geofabrik_json.geofabrik_overview.items():
                regionname = region
                try:
                    parent = value['parent']
                except KeyError:
                    parent = ''
                rurl = value['pbf_url']
                rshape = shape(value['geometry'])

                if parent in geofabrik_regions and regionname not in block_download:
                    # processing a country and no special sub-region
                    # check if rshape is subset of desired region. If so discard it
                    if wanted_region_polygon.contains(rshape):
                        # print (f'\t{regionname} is a subset of {wanted_map}, discard it')
                        continue
                    # check if rshape is a superset of desired region. if so discard it
                    if rshape.contains(wanted_region_polygon):
                        # print (f'\t{regionname} is a superset of {wanted_map}, discard it')
                        continue
                    # Check if rshape is a part of the tile / XY
                    if rshape.intersects(poly):
                        # print(f'\tintersecting tile: {regionname} tile={tile}')
                        if regionname not in must_download_maps:
                            must_download_maps.append(regionname)
                            must_download_urls.append(rurl)

        # If this tile contains the desired region, add it to the output
        # print (f'map= {wanted_map}\tmust_download= {must_download_maps}\tparent_added= {parent_added}\tforce_added= {force_added}')
        # if wanted_map in must_download_maps or parent_added == 1 or force_added == 1 or xy_mode is True:
        # first replace any forward slashes with underscores (us/texas to us_texas)
        must_download_maps = [sub.replace(
            '/', '_') for sub in must_download_maps]
        output.append({'x': tile['x'], 'y': tile['y'], 'left': tile['tile_left'], 'top': tile['tile_top'],
                       'right': tile['tile_right'], 'bottom': tile['tile_bottom'], 'countries': must_download_maps, 'urls': must_download_urls})

        return output

    def compose_bouding_box(self, xy_combination):
        """Overrides InformalGeofabrikInterface.calc_bouding_box()"""

        return {'top_x': xy_combination["x"], 'bot_x': xy_combination["x"],
                'top_y': xy_combination["y"], 'bot_y': xy_combination["y"]}

    def compose_shape(self, bbox_tiles):
        coords = [(bbox_tiles[0]["tile_top"], bbox_tiles[0]["tile_left"]), (bbox_tiles[0]["tile_top"], bbox_tiles[0]["tile_right"]),
                  (bbox_tiles[0]["tile_bottom"], bbox_tiles[0]["tile_right"]), (bbox_tiles[0]["tile_bottom"], bbox_tiles[0]["tile_left"])]
        coords_polygon = Polygon(coords)
        wanted_region = shape(coords_polygon)
        return wanted_region


def calc_bounding_box_tiles(bbox):
    bbox_tiles = []
    for x_value in range(bbox['top_x'], bbox['bot_x'] + 1):
        for y_value in range(bbox['top_y'], bbox['bot_y'] + 1):
            (tile_top, tile_left) = num2deg(x_value, y_value)
            (tile_bottom, tile_right) = num2deg(x_value+1, y_value+1)
            if tile_left < -180:
                tile_left = -180
            if tile_left > 180:
                tile_left = 180
            if tile_right < -180:
                tile_right = -180
            if tile_right > 180:
                tile_right = 180
            if tile_top < -90:
                tile_top = -90
            if tile_top > 90:
                tile_top = 90
            if tile_bottom < -90:
                tile_bottom = -90
            if tile_bottom > 90:
                tile_bottom = 90
            bbox_tiles.append({'x': x_value, 'y': y_value, 'tile_left': tile_left,
                               'tile_top': tile_top, 'tile_right': tile_right,
                               'tile_bottom': tile_bottom})

    return bbox_tiles


def deg2num(lat_deg, lon_deg, zoom=8):
    """
    Convert on./lat. to tile numbers
    """
    lat_rad = math.radians(lat_deg)
    n = 2.0 ** zoom
    xtile = int((lon_deg + 180.0) / 360.0 * n)
    ytile = int((1.0 - math.asinh(math.tan(lat_rad)) / math.pi) / 2.0 * n)
    return (xtile, ytile)


def num2deg(xtile, ytile, zoom=8):
    """
    Convert tile numbers to lon./lat.
    """
    n = 2.0 ** zoom
    lon_deg = xtile / n * 360.0 - 180.0
    lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / n)))
    lat_deg = math.degrees(lat_rad)
    return (lat_deg, lon_deg)


def get_countries_from_input(input_countries):
    """
    extract/split x/y combinations by given X/Y coordinates.
    input should be "188/88" or for multiple values "188/88,100/10,109/99".
    returns a list of x/y combinations as integers
    """

    countries = []

    # split by "," first for multiple x/y combinations, then by "/" for x and y value
    for country in input_countries.split(","):
        countries.append(country)

    return countries


def get_xy_coordinates_from_input(input_xy_coordinates):
    """
    extract/split x/y combinations by given X/Y coordinates.
    input should be "188/88" or for multiple values "188/88,100/10,109/99".
    returns a list of x/y combinations as integers
    """

    xy_combinations = []

    # split by "," first for multiple x/y combinations, then by "/" for x and y value
    for xy_coordinate in input_xy_coordinates.split(","):
        splitted = xy_coordinate.split("/")

        if len(splitted) == 2:
            xy_combinations.append(
                {"x": int(splitted[0]), "y": int(splitted[1])})

    return xy_combinations



"""
tests for the osm maps file
"""
import os
# import sys
import unittest

# import custom python packages
from wahoomc.osm_maps_functions import OsmData
from wahoomc.osm_maps_functions import OsmMaps
# from wahoomc.osm_maps_functions import TileNotFoundError
from wahoomc.input import InputData
from wahoomc import file_directory_functions as fd_fct
from wahoomc import constants


class TestOsmMapsCalculation(unittest.TestCase):
    """
    tests for the OSM maps file
    """

    # def setUp(self):

    def test_calc_border_countries_input_country(self):
        """
        Test initialized border countries
        - of malta
        - of germany
        """

        # malta
        self.process_and_check_border_countries(
            'malta', True, {'italy': {}, 'malta': {}}, 'country')

        # germany
        expected_result = {'czech-republic': {}, 'germany': {}, 'austria': {}, 'liechtenstein': {},
                           'switzerland': {}, 'italy': {}, 'netherlands': {}, 'belgium': {},
                           'luxembourg': {}, 'france': {}, 'poland': {}, 'denmark': {}, 'sweden': {}}
        self.process_and_check_border_countries(
            'germany', True, expected_result, 'country')

    def test_calc_border_countries_input_xy_coordinates_1tile(self):
        """
        Test initialized border countries
        - of a file with 1 tile and two countries
        """

        # one tile - france and germany
        expected_result = {'france': {}, 'germany': {}}
        self.process_and_check_border_countries(
            "133/88", True, expected_result, 'xy_coordinate')

    def test_calc_border_countries_input_xy_coordinates_2tiles(self):
        """
        Test initialized border countries
        - of a file with 2 tiles and one country
        """

        # two tiles - germany
        expected_result = {'germany': {}}
        self.process_and_check_border_countries(
            "134/87,134/88", True, expected_result, 'xy_coordinate')

    def test_calc_without_border_countries_input_country(self):
        """
        Test initialized countries without border countries
        - of germany
        - of china
        - of one tile
        """

        # germany
        self.process_and_check_border_countries(
            'germany', False, {'germany': {}}, 'country')

        # china
        self.process_and_check_border_countries(
            'china', False, {'china': {}}, 'country')

    def test_calc_without_border_countries__xy_coordinates_1tile(self):
        """
        Test initialized countries without border countries
        - of one tile
        """

        # one tile - france and germany
        expected_result = {'france': {}, 'germany': {}}
        self.process_and_check_border_countries(
            "133/88", False, expected_result, 'xy_coordinate')

    def test_calc_without_border_countries__xy_coordinates_2tiles(self):
        """
        Test initialized countries without border countries
        - of two tiles
        """

        # two tiles - germany
        expected_result = {'germany': {}}
        self.process_and_check_border_countries(
            "134/87,134/88", False, expected_result, 'xy_coordinate')

    def process_and_check_border_countries(self, inp_val, calc_border_c, exp_result, inp_mode):
        """
        helper method to process a country or json file and check the calculated border countries
        """

        o_input_data = InputData()
        if inp_mode == 'country':
            o_input_data.country = inp_val
        elif inp_mode == 'xy_coordinate':
            o_input_data.xy_coordinates = inp_val
        o_input_data.process_border_countries = calc_border_c

        o_osm_data = OsmData()
        o_osm_data.process_input_of_the_tool(o_input_data)

        result = o_osm_data.border_countries

        # delete the path to the file, here, only the correct border countries are checked
        for res in result:
            result[res] = {}

        self.assertEqual(result, exp_result)


class TestOSMMapsInput(unittest.TestCase):
    """
    tests for input of OsmData
    """

    def test_input_country_malta(self):
        """
        Test "malta" as input to the wahooMapsCreator
        check, if the given input-parameter is saved to the OsmMaps instance
        """

        o_input_data = InputData()
        o_input_data.country = 'malta'

        o_osm_data = OsmData()
        o_osm_data.process_input_of_the_tool(o_input_data)

        result = o_osm_data.country_name
        self.assertEqual(result, 'malta')

    def test_encoding_open_sea_osm(self):
        """
        use static json files in the repo to calculate relevant tile
        """

        with open(os.path.join(constants.RESOURCES_DIR, 'sea.osm')) as sea_file:  # pylint: disable=unspecified-encoding
            sea_data_no_encoding = sea_file.read()
        with open(os.path.join(constants.RESOURCES_DIR, 'sea.osm'), encoding="utf-8") as sea_file:
            sea_data_utf8 = sea_file.read()

        self.assertEqual(sea_data_no_encoding, sea_data_utf8)


class TestConfigFile(unittest.TestCase):
    """
    tests for the config .json file in the "wahooMapsCreatorData/_tiles/{country}" directory
    """

    def test_version_and_tags_of_country_config_file(self):
        """
        tests, if the return value of version is OK and if the tags are the same
        """

        o_input_data = InputData()
        o_input_data.country = 'malta'
        # prevent from downloading land_polygons each time
        o_input_data.max_days_old = 1000

        o_osm_data = OsmData()
        o_downloader = o_osm_data.process_input_of_the_tool(o_input_data)

        # download files marked for download to fill up map_file per country to write to config
        o_downloader.download_files_if_needed()

        o_osm_maps = OsmMaps(o_osm_data)

        o_osm_maps.write_country_config_file(o_input_data.country)

        self.assertTrue(
            o_osm_maps.tags_are_identical_to_last_run(o_input_data.country))

        country_config = fd_fct.read_json_file(os.path.join(
            constants.USER_OUTPUT_DIR, o_input_data.country, ".config.json"))

        self.assertEqual(constants.VERSION, country_config["version_last_run"])


if __name__ == '__main__':
    unittest.main()



"""
tests for setup functions
"""
import unittest
import platform
import os
import pkg_resources

# import custom python packages
from wahoomc.setup_functions import is_program_installed, is_map_writer_plugin_installed, \
    read_version_last_run
from wahoomc.constants_functions import get_tooling_win_path
from wahoomc.constants import USER_WAHOO_MC, VERSION
from wahoomc.file_directory_functions import write_json_file_generic
from wahoomc.downloader import get_latest_pypi_version


class TestSetup(unittest.TestCase):
    """
    tests for the required non-python programs of the python module
    """

    def test_installed_programs_mac(self):
        """
        tests, if the mac-relevant programs are installed
        """
        if platform.system() != "Windows":
            self.check_installation_of_program("java")

            self.check_installation_of_program("osmium")
            self.check_installation_of_program("osmosis")

            result = is_map_writer_plugin_installed()
            self.assertTrue(result)

    def test_installed_programs_windows(self):
        """
        tests, if the windows-relevant programs are installed
        """
        if platform.system() == "Windows":
            self.check_installation_of_program("java")

            self.assertTrue(os.path.exists(get_tooling_win_path(
                os.path.join('Osmosis', 'bin', 'osmosis.bat'), in_user_dir=True)))
            self.assertTrue(os.path.exists(
                get_tooling_win_path('osmconvert.exe')))
            self.assertTrue(os.path.exists(
                get_tooling_win_path('osmfilter.exe', in_user_dir=True)))
            self.assertTrue(os.path.exists(get_tooling_win_path('7za.exe')))

    def check_installation_of_program(self, program):
        """
        tests, if a given program is installed
        """
        result = is_program_installed(program)

        self.assertTrue(result)


class TestConfigFile(unittest.TestCase):
    """
    tests for the config .json file in the "wahooMapsCreatorData" directory
    """

    config_file_path = os.path.join(USER_WAHOO_MC, ".config.json")

    def test_version_if_no_config_file_exists(self):
        """
        tests, if the return value is None if the config file is deleted
        """
        if os.path.exists(self.config_file_path):
            os.remove(self.config_file_path)

        self.assertEqual(None, read_version_last_run())

    def test_version_if_written_eq(self):
        """
        tests, if return version is the one saved
        """
        # write dictionary to config file
        write_json_file_generic(self.config_file_path, {
                                "version_last_run": '2.0.2'})

        self.assertEqual('2.0.2', read_version_last_run())

    def test_version_if_written_neq(self):
        """
        tests, if the comparison of the returned version_last_run is OK
        """
        # write dictionary to config file
        write_json_file_generic(self.config_file_path, {
                                "version_last_run": '2.0.3'})

        self.assertNotEqual('2.0.2', read_version_last_run())

    def test_version_constants_against_pypi(self):
        """
        tests, if the version of constants.py is equal to the latest available version on PyPI
        """
        latest_version = pkg_resources.parse_version(
            get_latest_pypi_version()).public

        self.assertEqual(
            VERSION, latest_version)


if __name__ == '__main__':
    unittest.main()



"""
Object for accessing Geofabrik .json file
"""
#!/usr/bin/python

# import official python packages
import geojson  # pylint: disable=import-error

# import custom python packages
from wahoomc.constants import GEOFABRIK_PATH


class CountyIsNoGeofabrikCountry(Exception):
    """Raised when actual country is not a geofabrik country"""


class GeofabrikJson:
    """
    This is a Geofabrik .json processing class for constants in the Geofabrik .json file
    """
    raw_json = None
    geofabrik_overview = {}
    geofabrik_region_overview = {}
    geofabrik_regions = []

    def __init__(self):
        # read geofabrik .json file and fill class-attributes // access is only once
        if GeofabrikJson.raw_json is None and not GeofabrikJson.geofabrik_overview \
                and not GeofabrikJson.geofabrik_region_overview and not GeofabrikJson.geofabrik_regions:
            GeofabrikJson.raw_json, GeofabrikJson.geofabrik_overview, GeofabrikJson.geofabrik_region_overview, GeofabrikJson.geofabrik_regions = self.read_geofabrik_json_file()

    def read_geofabrik_json_file(self):
        """
        read geofabrik .json file and fill class-attributes

        geofabrik_regions as i defined them are the ones without parent
        geofabrik_overview contain all entries, with and without parent
        """
        raw_json = []
        geofabrik_overview = {}
        geofabrik_region_overview = {}
        geofabrik_regions = []

        with open(GEOFABRIK_PATH, encoding='utf8') as file_handle:
            raw_json = geojson.load(file_handle)
        file_handle.close()

        # create a dict with information easy to access because they are often needed
        for feature in raw_json.features:
            props = feature.properties
            id_no = props['id']
            pbf_url = props['urls']['pbf']

            try:
                parent = props['parent']

                geofabrik_overview[id_no] = {
                    'parent': parent,
                    'pbf_url': pbf_url,
                    'geometry': feature.geometry}
            except KeyError:
                geofabrik_overview[id_no] = {
                    'pbf_url': pbf_url,
                    'geometry': feature.geometry}
                geofabrik_region_overview[id_no] = {
                    'pbf_url': pbf_url}
                geofabrik_regions.append(id_no)

        return raw_json, geofabrik_overview, geofabrik_region_overview, geofabrik_regions

    def get_geofabrik_parent_country(self, id_no):
        """
        Get the parent map/region of a region from the already loaded json data
        """
        id_no_translated = self.translate_id_no_to_geofabrik(id_no)

        try:
            entry = self.geofabrik_overview[id_no_translated]
            if 'parent' in entry:
                return (entry['parent'], id_no_translated)

            return ('', id_no_translated)
        except KeyError:
            return None, None

    def get_geofabrik_url(self, id_no):
        """
        Get the map download url from a region with the already loaded json data
        """
        id_no_translated = self.translate_id_no_to_geofabrik(id_no)
        try:
            entry = self.geofabrik_overview[id_no_translated]
            if 'pbf_url' in entry:
                return entry['pbf_url']
        except KeyError:
            pass

        return None

    def get_geofabrik_geometry(self, id_no):
        """
        Get the geometry from a region with the already loaded json data
        """
        id_no_translated = self.translate_id_no_to_geofabrik(id_no)
        try:
            entry = self.geofabrik_overview[id_no_translated]
            if 'geometry' in entry:
                return entry['geometry']
        except KeyError:
            pass

        return None

    def is_input_a_geofabrik_id_no(self, id_no):
        """
        check if the given input is a geofabrik id number
        """
        if id_no in self.geofabrik_overview:
            return True

        return False

    def translate_id_no_to_geofabrik(self, country):
        """
        get geofabrik id by country .json filename
        """

        if country in self.geofabrik_overview:
            return country

        if country.replace('_', '-') in self.geofabrik_overview:
            return country.replace('_', '-')

        if 'us/'+country.replace('_', '-') in self.geofabrik_overview:
            return 'us/'+country.replace('_', '-')

        # if none of them got triggert --> exception
        raise CountyIsNoGeofabrikCountry



#!/usr/bin/python3
# pylint: skip-file

"""
This script is designed to act as assistance in converting shapefiles
to OpenStreetMap data. This file is optimized and tested with MassGIS
shapefiles, converted to EPSG:4326 before being passed to the script.
You can perform this conversion with 

   ogr2ogr -t_srs EPSG:4326 new_file.shp old_file.shp

It is expected that you will modify the fixed_tags, tag_mapping, and
boring_tags attributes of this script before running. You should read,
or at least skim, the code up until it says:

  DO NOT CHANGE AFTER THIS LINE.

to accomodate your own data. 

Modified to support a maximum waylength and multiple outer ways in a
multipolygon relationship

adapted by devemux86
"""

import sys
__author__ = "Christopher Schmidt <crschmidt@crschmidt.net>"
__version__ = "$Id: polyshp2osm-WaterBody.py,v 1.3 2009/03/30 20:54:13 tbook Exp $"

gdal_install = """
Installing GDAL depends on your platform. Information is available at:
   
   http://trac.osgeo.org/gdal/wiki/DownloadingGdalBinaries

For Debian-based systems:

   apt-get install python-gdal

will usually suffice. 
"""

# These tags are attached to all exterior ways. You can put any key/value pairs
# in this dictionary.

fixed_tags = {
    'natural': 'nosea',
    'layer': '-5'
}

# Here are a number of functions: These functions define tag mappings. The API
# For these functions is that they are passed the attributes from a feature,
# and they return a list of two-tuples which match to key/value pairs.

# The following Ftypes are not imported
# 46100 - Submerged Stream
ignoreField = "tile_x"
ignoreValues = []

# The following Ftypes are imported, but don't have any OSM tags attached:
# 36100 - Playa - An area from which water evaporates


def ftype(data):
    """Type of body - From NHD Ftype"""
    natural = {
        'LakePond': 'water',
        'SwampMarsh': 'wetland'
    }

    if 'ftype' in data:
        if data['ftype'] in natural:
            return [('natural', natural[data['ftype']])]
    return None


def fcode(data):
    """For features where the ftype is not specific enough"""
    landuse = {
        '43600': 'reservoir',  # Reservoir
        '43601': 'reservoir',  # Reservoir
        '43602': 'reservoir',  # Reservoir
        '43603': 'reservoir',  # Reservoir
        '43604': 'reservoir',  # Reservoir
        '43605': 'reservoir',  # Reservoir
        '43606': 'reservoir',  # Reservoir
        '43607': 'reservoir',  # Reservoir
        '43609': 'reservoir',  # Reservoir
        '43610': 'reservoir',  # Reservoir
        '43611': 'reservoir',  # Reservoir
        '43612': 'reservoir',  # Reservoir
        '43614': 'reservoir',  # Reservoir
        '43615': 'reservoir',  # Reservoir
        '43616': 'reservoir',  # Reservoir
        '43617': 'reservoir',  # Reservoir
        '43618': 'reservoir',  # Reservoir
        '43619': 'reservoir',  # Reservoir
        '43620': 'reservoir'  # Reservoir
    }

    manmade = {
        '43613': 'reservoir_covered'  # Covered Reservoir
    }

    leisure = {
        '43608': 'swimming_pool'  # Swimming Pool
    }

    if 'fcode' in data:
        keys = []
        if data['fcode'] in landuse:
            keys.append(('landuse', landuse[data['fcode']]))
        if data['fcode'] in manmade:
            keys.append(('man_made', manmade[data['fcode']]))
        if data['fcode'] in leisure:
            keys.append(('leisure', leisure[data['fcode']]))
        return keys
    return None

# The most important part of the code: define a set of key/value pairs
# to iterate over to generate keys. This is a list of two-tuples: first
# is a 'key', which is only used if the second value is a string. In
# that case, it is a map of lowercased fieldnames to OSM tag names: so
# fee_owner maps to 'owner' in the OSM output.

# if the latter is callable (has a __call__; is a function), then that
# method is called, passing in a dict of feature attributes with
# lowercased key names. Those functions can then return a list of
# two-tuples to be used as tags, or nothin' to skip the tags.


tag_mapping = [
    ('ftype', ftype),
    ('fcode', fcode),
    ('gnis_name', 'name'),
    ('gnis_id', 'gnis:feature_id'),
    ('elevation', 'ele'),
]

# These tags are not exported, even with the source data; this should be
# used for tags which are usually calculated in a GIS. AREA and LEN are
# common.

boring_tags = ['error', 'tile_x', 'tile_y']

# Namespace is used to prefix existing data attributes. If 'None', or
# '--no-source' is set, then source attributes are not exported, only
# attributes in tag_mapping.

namespace = "NHD"
#namespace = None

# Uncomment the "DONT_RUN = False" line to get started.

#DONT_RUN = True
DONT_RUN = False

# Set the maximum length of a way (in nodes) before it is split into
# shorter ways

Max_Waylength = 1500000

# =========== DO NOT CHANGE AFTER THIS LINE. ===========================
# Below here is regular code, part of the file. This is not designed to
# be modified by users.
# ======================================================================


try:
    try:
        from osgeo import ogr
    except ImportError:
        import ogr
except ImportError:
    __doc__ += gdal_install
    if DONT_RUN:
        print(__doc__)
        sys.exit(2)
    print("OGR Python Bindings not installed.\n%s" % gdal_install)
    sys.exit(1)


def close_file():
    """ Internal. Close an open file."""
    global open_file
    if not open_file.closed:
        open_file.write("</osm>")
        open_file.close()


def start_new_file():
    """ Internal. Open a new file, closing existing file if neccesary."""
    global open_file, file_counter
    file_counter += 1
    if open_file:
        close_file()
    open_file = open("%s%s.osm" % (file_name, file_counter), "w")
    print("<osm version='0.6'>", file=open_file)


def clean_attr(val):
    """Internal. Hacky way to make attribute XML safe."""
    val = str(val)
    val = val.replace("&", "&amp;").replace("'", "&quot;").replace(
        "<", "&lt;").replace(">", "&gt;").strip()
    return val


def add_ring_nodes(ring):
    """Internal. Write the outer ring nodes."""
    global open_file, id_counter
    ringways = []
    ids = []
    firstnode = id_counter
    if range(ring.GetPointCount() - 1) == 0 or ring.GetPointCount() == 0:
        print("Degenerate ring.", file=sys.stderr)
        return
    for count in range(ring.GetPointCount() - 1):
        ids.append(id_counter)
        print("<node timestamp='1969-12-31T23:59:59Z' changeset='-1' id='%s' version='1' lon='%s' lat='%s' />" %
              (id_counter, ring.GetX(count), ring.GetY(count)), file=open_file)
        id_counter += 1
        if (count > 0) and ((count % (Max_Waylength - 1)) == 0):
            ringways.append(ids)
            ids = []
            ids.append(id_counter - 1)
    ids.append(firstnode)
    ringways.append(ids)
    return ringways


def add_ring_way(ring):
    """Internal. write out the 'holes' in a polygon."""
    global open_file, id_counter
    ids = []
    ringways = []
    for count in range(ring.GetPointCount() - 1):
        ids.append(id_counter)
        print("<node timestamp='1969-12-31T23:59:59Z' changeset='-1' version='1' id='%s' lon='%s' lat='%s' />" %
              (id_counter, ring.GetX(count), ring.GetY(count)), file=open_file)
        id_counter += 1
    if len(ids) == 0:
        return None
    print("<way timestamp='1969-12-31T23:59:59Z' changeset='-1' id='%s' version='1'>" %
          id_counter, file=open_file)
    way_id = id_counter
    id_counter += 1
    count = 0
    for i in ids:
        print("<nd ref='%s' />" % i, file=open_file)
        count += 1
        if count >= Max_Waylength - 1:
            count = 0
            print("</way>", file=open_file)
            ringways.append(way_id)
            print("<way timestamp='1969-12-31T23:59:59Z' changeset='-1'id='%s' version='1'>" %
                  id_counter, file=open_file)
            way_id = id_counter
            id_counter += 1
            print("<nd ref='%s' />" % i, file=open_file)
    print("<nd ref='%s' />" % ids[0], file=open_file)
    print("</way>", file=open_file)
    ringways.append(way_id)

    return ringways


open_file = None

file_name = None

id_counter = 22951459320

file_counter = 0
counter = 0


class AppError(Exception):
    pass


def run(filename, slice_count=1, obj_count=100000000, output_location=None, no_source=False):
    """Run the converter. Requires open_file, file_name, id_counter,
    file_counter, counter to be defined in global space; not really a very good
    singleton."""
    global id_counter, file_counter, counter, file_name, open_file, namespace

    if no_source:
        namespace = None

    if output_location:
        file_name = output_location

    ds = ogr.Open(filename)
    if not ds:
        raise AppError("OGR Could not open the file %s" % filename)
    l = ds.GetLayer(0)

    max_objs_per_file = obj_count

    extent = l.GetExtent()
    if extent[0] < -180 or extent[0] > 180 or extent[2] < -90 or extent[2] > 90:
        raise AppError("Extent does not look like degrees; are you sure it is? \n(%s, %s, %s, %s)" % (
            extent[0], extent[2], extent[1], extent[3]))
    slice_width = (extent[1] - extent[0]) / slice_count

    seen = {}

    print("Running %s slices with %s base filename against shapefile %s" % (
        slice_count, file_name, filename))

    for i in range(slice_count):

        l.ResetReading()
        l.SetSpatialFilterRect(
            extent[0] + slice_width * i, extent[2], extent[0] + (slice_width * (i + 1)), extent[3])

        start_new_file()
        f = l.GetNextFeature()

        obj_counter = 0
        last_obj_split = 0

        while f:
            start_id_counter = id_counter
            if f.GetFID() in seen:
                f = l.GetNextFeature()
                continue
            seen[f.GetFID()] = True

#            done = False
#            while f.GetField(ignoreField) in ignoreValues:
#                if l.GetNextFeature() == None:
#                    done = True
#                    break
#                f = l.GetNextFeature()
#            if done == True:
#                break

            outerways = []
            innerways = []

            geom = f.GetGeometryRef()
            ring = geom.GetGeometryRef(0)

            objcount = ring.GetPointCount()
            for i in range(1, geom.GetGeometryCount()):
                objcount += geom.GetGeometryRef(i).GetPointCount()
                objcount += 1

            if (obj_counter - last_obj_split + objcount) > max_objs_per_file:
                print("Splitting file with %s objs" %
                      (obj_counter - last_obj_split))
                start_new_file()
                last_obj_split = obj_counter

            if objcount > max_objs_per_file:
                print("Warning: a feature contains %i objects which is more than the %i object limit.  It will be placed in a file by itself." % (
                    objcount, max_objs_per_file))

            ringways = add_ring_nodes(ring)
            if not ringways or len(ringways) == 0:
                f = l.GetNextFeature()
                continue
# Write out the outer ways in the relation
            for count in range(len(ringways)):
                ids = ringways[count]
                if ids and len(ids) > 1:
                    print("<way timestamp='1969-12-31T23:59:59Z' changeset='-1' id='%s' version='1'>" %
                          id_counter, file=open_file)
                    outerways.append(id_counter)
                    id_counter += 1
                    for i in ids:
                        print("<nd ref='%s' />" % i, file=open_file)
# Write out the fields for the way
                    field_count = f.GetFieldCount()
                    fields = {}
                    for field in range(field_count):
                        value = f.GetFieldAsString(field)
                        name = f.GetFieldDefnRef(field).GetName()
                        if namespace and name and value and name not in boring_tags:
                            print("<tag k='%s:%s' v='%s' />" % (namespace,
                                  name, clean_attr(value)), file=open_file)
                        fields[name.lower()] = value
                    tags = {}
# Perform the specified field mappting
                    for tag_name, map_value in tag_mapping:
                        if hasattr(map_value, '__call__'):
                            tag_values = map_value(fields)
                            if tag_values:
                                for tag in tag_values:
                                    tags[tag[0]] = tag[1]
                        else:
                            if tag_name in fields:
                                tags[map_value] = fields[tag_name].title()
                    for key, value in tags.items():
                        if key and value:
                            print("<tag k='%s' v='%s' />" %
                                  (key, clean_attr(value)), file=open_file)
# Write fixed tabs
                    for name, value in fixed_tags.items():
                        print("<tag k='%s' v='%s' />" %
                              (name, clean_attr(value)), file=open_file)
                    print("</way>", file=open_file)
            if (geom.GetGeometryCount() > 1) or (len(ringways) > 1):
                # add the inner ways
                for i in range(1, geom.GetGeometryCount()):
                    ringways = add_ring_way(geom.GetGeometryRef(i))
                    for way in ringways:
                        innerways.append(way)
                print("<relation timestamp='1969-12-31T23:59:59Z' changeset='-1' id='%s' version='1' ><tag k='type' v='multipolygon' />" %
                      id_counter, file=open_file)
                id_counter += 1
                for way in outerways:
                    print('<member type="way" ref="%s" role="outer" />' %
                          way, file=open_file)
                for way in innerways:
                    print('<member type="way" ref="%s" role="inner" />' %
                          way, file=open_file)
                print("</relation>", file=open_file)

            counter += 1
            f = l.GetNextFeature()
            obj_counter += (id_counter - start_id_counter)

        close_file()


if __name__ == "__main__":
    if DONT_RUN:
        print(__doc__)
        sys.exit(2)

    from optparse import OptionParser

    parse = OptionParser(
        usage="%prog [args] filename.shp", version=__version__)
    parse.add_option("-s", "--slice-count", dest="slice_count",
                     help="Number of horizontal slices of data", default=1,
                     action="store", type="int")
    parse.add_option("-o", "--obj-count",
                     dest="obj_count",
                     help="Maximum number of objects in a single .osm file",
                     default=1000000, type="int")
    parse.add_option("-n", "--no-source", dest="no_source",
                     help="Do not store source attributes as tags.",
                     action="store_true", default=False)
    parse.add_option("-l", "--output-location",
                     dest="output_location", help="base filepath for output files.",
                     default="poly_output")
    (options, args) = parse.parse_args()

    if not len(args):
        print("No shapefile name given!")
        parse.print_help()
        sys.exit(3)

    kw = {}
    for key in ('slice_count', 'obj_count', 'output_location', 'no_source'):
        kw[key] = getattr(options, key)

    try:
        run(args[0], **kw)
    except AppError as E:
        print("An error occurred: \n%s" % E)



"""
tests for the downloader file
"""
import os
import unittest
import mock


from wahoomc.constants_functions import translate_tags_to_keep, \
    get_tag_wahoo_xml_path, TagWahooXmlNotFoundError
from wahoomc.constants import RESOURCES_DIR


tags_universal_simple = {"TAGS_TO_KEEP_UNIVERSAL": {
    'access': '',
    'area': 'yes'
}}

tags_universal_adv = {"TAGS_TO_KEEP_UNIVERSAL": {
    'access': '',
    'area': 'yes',
    'bicycle': '',
    'bridge': '',
    'foot': ['ft_yes', 'foot_designated']
}}


class TestTranslateTags(unittest.TestCase):
    """
    tests for translating tags-constants between the universal format and OS-specific formats
    """

    @ mock.patch("wahoomc.file_directory_functions.json.load")
    @ mock.patch("wahoomc.open")
    def test_translate_tags_to_keep_simple_macos(self, mock_open, mock_json_load):  # pylint: disable=unused-argument
        """
        Test translating tags to keep from universal format to macOS
        """
        tags = ['access', 'area=yes']
        mock_json_load.return_value = tags_universal_simple

        transl_tags = translate_tags_to_keep()
        self.assertEqual(tags, transl_tags)

    @ mock.patch("wahoomc.file_directory_functions.json.load")
    @ mock.patch("wahoomc.open")
    def test_translate_tags_to_keep_simple_win(self, mock_open, mock_json_load):  # pylint: disable=unused-argument
        """
        Test translating tags to keep from universal format to Windows
        """
        tags_win = 'access= area=yes'
        mock_json_load.return_value = tags_universal_simple

        transl_tags = translate_tags_to_keep(sys_platform='Windows')
        self.assertEqual(tags_win, transl_tags)

    @ mock.patch("wahoomc.file_directory_functions.json.load")
    @ mock.patch("wahoomc.open")
    def test_translate_tags_to_keep_adv_macos(self, mock_open, mock_json_load):  # pylint: disable=unused-argument
        """
        Test translating tags to keep from universal format to macOS
        """
        tags = ['access', 'area=yes', 'bicycle',
                'bridge', 'foot=ft_yes, foot_designated']
        mock_json_load.return_value = tags_universal_adv

        transl_tags = translate_tags_to_keep()
        self.assertEqual(tags, transl_tags)

    @ mock.patch("wahoomc.file_directory_functions.json.load")
    @ mock.patch("wahoomc.open")
    def test_translate_tags_to_keep_adv_win(self, mock_open, mock_json_load):  # pylint: disable=unused-argument
        """
        Test translating tags to keep from universal format to Windows
        """
        tags_win = 'access= area=yes bicycle= bridge= foot=ft_yes =foot_designated'
        mock_json_load.return_value = tags_universal_adv

        transl_tags = translate_tags_to_keep(sys_platform='Windows')
        self.assertEqual(tags_win, transl_tags)

    def test_translate_tags_to_keep_full_macos(self):
        """
        Test translating tags to keep from universal format to macOS // all "tags to keep"
        """
        tags = ['access', 'area=yes', 'bicycle', 'bridge', 'foot=ft_yes, foot_designated',
                'amenity=fuel, cafe, drinking_water', 'shop=bakery, bicycle',
                'highway=abandoned, bus_guideway, disused, bridleway, byway, construction, cycleway, footway, living_street, motorway, motorway_link, path, pedestrian, primary, primary_link, residential, road, secondary, secondary_link, service, steps, tertiary, tertiary_link, track, trunk, trunk_link, unclassified',
                'natural=coastline, nosea, sea, beach, land, scrub, water, wetland, wood',
                'landuse=forest, commercial, industrial, residential, retail',
                'leisure=park, nature_reserve', 'railway=rail, tram, station, stop',
                'surface', 'tracktype', 'tunnel', 'waterway=canal, drain, river, riverbank', 'wood=deciduous']

        transl_tags = translate_tags_to_keep(use_repo=True)
        self.assertEqual(tags, transl_tags)

    def test_translate_tags_to_keep_full_win(self):
        """
        Test translating tags to keep from universal format to Windows // all "tags to keep"
        """
        tags_win = 'access= area=yes bicycle= bridge= foot=ft_yes =foot_designated amenity=fuel =cafe =drinking_water shop=bakery =bicycle highway=abandoned =bus_guideway =disused =bridleway =byway =construction =cycleway =footway =living_street =motorway =motorway_link =path =pedestrian =primary =primary_link =residential =road =secondary =secondary_link =service =steps =tertiary =tertiary_link =track =trunk =trunk_link =unclassified natural=coastline =nosea =sea =beach =land =scrub =water =wetland =wood landuse=forest =commercial =industrial =residential =retail leisure=park =nature_reserve railway=rail =tram =station =stop surface= tracktype= tunnel= waterway=canal =drain =river =riverbank wood=deciduous'

        transl_tags = translate_tags_to_keep(
            sys_platform='Windows', use_repo=True)
        self.assertEqual(tags_win, transl_tags)

    def test_translate_name_tags_to_keep_full_macos(self):
        """
        Test translating name tags to keep from universal format to Windows // all "name tags to keep"
        """
        names_tags = ['admin_level=2', 'area=yes', 'mountain_pass', 'natural',
                      'place=city, hamlet, island, isolated_dwelling, islet, locality, suburb, town, village, country']

        transl_tags = translate_tags_to_keep(name_tags=True, use_repo=True)
        self.assertEqual(names_tags, transl_tags)

    def test_translate_name_tags_to_keep_full_win(self):
        """
        Test translating name tags to keep from universal format to macOS // all "name tags to keep"
        """

        names_tags_win = 'admin_level=2 area=yes mountain_pass= natural= place=city =hamlet =island =isolated_dwelling =islet =locality =suburb =town =village =country'

        transl_tags = translate_tags_to_keep(
            name_tags=True, sys_platform='Windows', use_repo=True)
        self.assertEqual(names_tags_win, transl_tags)


class TestTagWahooXML(unittest.TestCase):
    """
    tests for tag-wahoo xml file
    """

    def test_not_existing_tag_wahoo_xml(self):
        """
        check if a non-existing tag-wahoo xml file issues an exception
        """
        with self.assertRaises(TagWahooXmlNotFoundError):
            get_tag_wahoo_xml_path("not_existing.xml")

    def test_existing_tag_wahoo_xml(self):
        """
        check if the correct path of an existing tag-wahoo xml file is returned
        """

        expected_path = os.path.join(
            RESOURCES_DIR, "tag_wahoo_adjusted", "tag-wahoo-poi.xml")
        self.assertEqual(get_tag_wahoo_xml_path(
            "tag-wahoo-poi.xml"), expected_path)


if __name__ == '__main__':
    unittest.main()



"""
functions and object for managing OSM maps
"""
#!/usr/bin/python

# import official python packages
from datetime import datetime
import glob
import multiprocessing
import os
import subprocess
import sys
import platform
import shutil
import logging

# import custom python packages
from wahoomc.file_directory_functions import read_json_file, create_empty_directories, write_json_file_generic
from wahoomc.constants_functions import translate_tags_to_keep, \
    get_tooling_win_path, get_tag_wahoo_xml_path, TagWahooXmlNotFoundError

from wahoomc.constants import USER_WAHOO_MC
from wahoomc.constants import USER_OUTPUT_DIR
from wahoomc.constants import RESOURCES_DIR
from wahoomc.constants import LAND_POLYGONS_PATH
from wahoomc.constants import VERSION
from wahoomc.constants import OSMOSIS_WIN_FILE_PATH

from wahoomc.downloader import Downloader
from wahoomc.geofabrik import CountryGeofabrik, XYGeofabrik

log = logging.getLogger('main-logger')


class TileNotFoundError(Exception):
    """Raised when no tile is found for x/y combination"""


def run_subprocess_and_log_output(cmd, error_message, cwd=""):
    """
    run given cmd-subprocess and issue error message if wished
    """
    if not cwd:
        with subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as process:
            for line in iter(process.stdout.readline, b''):  # b'\n'-separated lines
                try:
                    log.debug('subprocess:%r', line.decode("utf-8").strip())
                except UnicodeDecodeError:
                    log.debug('subprocess:%r', line.decode("latin-1").strip())
    else:
        with subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=cwd) as process:
            for line in iter(process.stdout.readline, b''):  # b'\n'-separated lines
                try:
                    log.debug('subprocess:%r', line.decode("utf-8").strip())
                except UnicodeDecodeError:
                    log.debug('subprocess:%r', line.decode("latin-1").strip())

    if error_message and process.wait() != 0:  # 0 means success
        log.error(error_message)
        sys.exit()


def get_timestamp_last_changed(file_path):
    """
    returns the timestamp of the last-changed datetime of the given file
    """
    chg_time = os.path.getmtime(file_path)

    return datetime.fromtimestamp(chg_time).isoformat()


class OsmData():  # pylint: disable=too-few-public-methods
    """
    object with all internal parameters to process maps
    """

    def __init__(self):
        """
        xxx
        """
        self.force_processing = False
        self.tiles = []
        self.border_countries = {}
        self.country_name = ''

    def process_input_of_the_tool(self, o_input_data):
        """
        Process input: get relevant tiles and if border countries should be calculated
        The three primary inputs are giving by a separate value each and have separate processing:
        1. country name
        2. x/y combinations

        # Check for not existing or expired files. Mark for download, if dl is needed
        # - land polygons file
        # - .osm.pbf files

        steps in this function:
        1. take over input paramters (force_processing is changed in the function further down)
        2. check + download geofabrik file if geofabrik-processing
        3. calculate relevant tiles for map creation
        4. calculate border countries for map creation
        5. evaluate the country-name for folder cration during processing
        6. calculate if force_processing should be set to true
        """

        o_downloader = Downloader(
            o_input_data.max_days_old, o_input_data.force_download, self.border_countries)
        # takeover what is given by user first for force_processing
        self.force_processing = o_input_data.force_processing

        log.info('-' * 80)

        # geofabrik file
        if o_downloader.should_geofabrik_file_be_downloaded():
            self.force_processing = True
            o_downloader.download_geofabrik_file()

        # calc tiles
        if o_input_data.country:
            self.calc_tiles_country(o_input_data)
        elif o_input_data.xy_coordinates:
            self.calc_tiles_xy(o_input_data)

        # calc border countries
        log.info('-' * 80)
        if o_input_data.country:
            self.calc_border_countries_country(o_input_data)
        elif o_input_data.xy_coordinates:
            self.calc_border_countries()
        # log border countries when and when not calculated to output the processed country(s)
        self.log_border_countries()

        # calc country name
        if o_input_data.country:
            # country name is the input argument
            self.country_name = o_input_data.country
        elif o_input_data.xy_coordinates:
            self.calc_country_name_xy()

        # calc force processing
        # Check for not existing or expired files. Mark for download, if dl is needed
        o_downloader.check_land_polygons_file()
        o_downloader.check_osm_pbf_file()

        # If one of the files needs to be downloaded, reprocess all files
        if o_downloader.need_to_dl:
            self.force_processing = True

        return o_downloader

    def calc_tiles_country(self, o_input_data):
        """
        option 1: input a country as parameter, e.g. germany
        """
        log.info('# Input country: %s.', o_input_data.country)

        # use Geofabrik-URL to calculate the relevant tiles
        o_geofabrik = CountryGeofabrik(o_input_data.country)
        self.tiles = o_geofabrik.get_tiles_of_wanted_map()

    def calc_tiles_xy(self, o_input_data):
        """
        option 2: input a x/y combinations as parameter, e.g. 134/88  or 133/88,130/100
        """
        log.info(
            '# Input X/Y coordinates: %s.', o_input_data.xy_coordinates)

        o_geofabrik = XYGeofabrik(o_input_data.xy_coordinates)
        # find the tiles for  x/y combinations in the geofabrik json files
        self.tiles = o_geofabrik.get_tiles_of_wanted_map()

    def calc_border_countries_country(self, o_input_data):
        """
        calculate the border countries for the given tiles when input is a country
        - if CLI/GUI input by user
        """
        if o_input_data.process_border_countries:
            self.calc_border_countries()
        # set the to-be-processed country as border country
        else:
            self.border_countries[o_input_data.country] = {}

    def calc_border_countries(self):
        """
        calculate the border countries for the given tiles. i.e.
        - if CLI/GUI input by user
        - if processing x/y coordinates
        """
        log.info('# Determine involved/border countries')

        # Build list of countries needed
        for tile in self.tiles:
            for country in tile['countries']:
                if country not in self.border_countries:
                    self.border_countries[country] = {}

    def log_border_countries(self):
        """
        write calculated border countries/involved countries to log
        """
        for country in self.border_countries:
            log.info('+ Involved country: %s', country)

        # input can be only one country, if there are more than one,
        # border countries must be selected
        if len(self.border_countries) > 1:
            log.info('+ Border countries will be processed')

    def calc_country_name_xy(self):
        """
        country name is the X/Y combinations separated by minus
        >1 x/y combinations are separated by underscore
        """
        for tile in self.tiles:
            if not self.country_name:
                self.country_name = f'{tile["x"]}-{tile["y"]}'
            else:
                self.country_name = f'{self.country_name}_{tile["x"]}-{tile["y"]}'


class OsmMaps:
    """
    This is a OSM data class
    """

    # Number of workers for the Osmosis read binary fast function
    workers = '1'

    def __init__(self, o_osm_data):
        self.o_osm_data = o_osm_data
        self.osmconvert_path = get_tooling_win_path('osmconvert')

        create_empty_directories(
            USER_OUTPUT_DIR, self.o_osm_data.tiles, self.o_osm_data.border_countries)

    def filter_tags_from_country_osm_pbf_files(self):  # pylint: disable=too-many-statements
        """
        Filter tags from country osm.pbf files
        """

        log.info('-' * 80)
        log.info('# Filter tags from country osm.pbf files')

        for key, val in self.o_osm_data.border_countries.items():
            # evaluate contry directory, create if not exists
            country_dir = os.path.join(USER_OUTPUT_DIR, key)

            # set names for filtered files for WIN, later on add ".pbf" for macOS/Linux
            out_file_o5m_filtered_win = os.path.join(country_dir,
                                                     'filtered.o5m')
            out_file_o5m_filtered_names_win = os.path.join(country_dir,
                                                           'filtered_names.o5m')

            # Windows
            if platform.system() == "Windows":
                out_file_o5m = os.path.join(country_dir, 'outFile.o5m')
                # only create o5m file if not there already or force processing (no user input possible)
                # --> speeds up processing if one only wants to test tags / POIs
                if not os.path.isfile(out_file_o5m) or self.o_osm_data.force_processing is True \
                        or self.last_changed_is_identical_to_last_run(key) is False:
                    log.info('+ Converting map of %s to o5m format', key)
                    cmd = [self.osmconvert_path]
                    cmd.extend(['-v', '--hash-memory=2500', '--complete-ways',
                                '--complete-multipolygons', '--complete-boundaries',
                                '--drop-author', '--drop-version'])
                    cmd.append(val['map_file'])
                    cmd.append('-o='+out_file_o5m)

                    run_subprocess_and_log_output(
                        cmd, f'! Error in OSMConvert with country: {key}')
                else:
                    log.info('+ Map of %s already in o5m format', key)

                # filter out tags:
                # - if no filtered files exist
                # - force processing is set (this is also when new map files were dowwnloaded)
                # - the defined TAGS_TO_KEEP_UNIVERSAL constants have changed are changed (user input or new release)
                if not os.path.isfile(out_file_o5m_filtered_win) or not os.path.isfile(out_file_o5m_filtered_names_win) \
                        or self.o_osm_data.force_processing is True or self.tags_are_identical_to_last_run(key) is False \
                        or self.last_changed_is_identical_to_last_run(key) is False:
                    log.info(
                        '+ Filtering unwanted map objects out of map of %s', key)
                    cmd = [get_tooling_win_path('osmfilter', in_user_dir=True)]
                    cmd.append(out_file_o5m)
                    cmd.append(
                        '--keep="' + translate_tags_to_keep(sys_platform=platform.system()) + '"')
                    cmd.append('--keep-tags="all type= layer= ' +
                               translate_tags_to_keep(sys_platform=platform.system()) + '"')
                    cmd.append('-o=' + out_file_o5m_filtered_win)

                    run_subprocess_and_log_output(
                        cmd, f'! Error in OSMFilter with country: {key}')

                    cmd = [get_tooling_win_path('osmfilter', in_user_dir=True)]
                    cmd.append(out_file_o5m)
                    cmd.append(
                        '--keep="' + translate_tags_to_keep(
                            name_tags=True, sys_platform=platform.system()) + '"')
                    cmd.append('--keep-tags="all type= name= layer= ' +
                               translate_tags_to_keep(
                                   name_tags=True, sys_platform=platform.system()) + '"')
                    cmd.append('-o=' + out_file_o5m_filtered_names_win)

                    run_subprocess_and_log_output(
                        cmd, f'! Error in OSMFilter with country: {key}')

                val['filtered_file'] = out_file_o5m_filtered_win
                val['filtered_file_names'] = out_file_o5m_filtered_names_win

            # Non-Windows
            else:
                out_file_pbf_filtered_mac = f'{out_file_o5m_filtered_win}.pbf'
                out_file_pbf_filtered_names_mac = f'{out_file_o5m_filtered_names_win}.pbf'

                # filter out tags:
                # - if no filtered files exist
                # - force processing is set (this is also when new map files were dowwnloaded)
                # - the defined TAGS_TO_KEEP_UNIVERSAL constants have changed are changed (user input or new release)
                if not os.path.isfile(out_file_pbf_filtered_mac) or not os.path.isfile(out_file_pbf_filtered_names_mac) \
                        or self.o_osm_data.force_processing is True or self.tags_are_identical_to_last_run(key) is False \
                        or self.last_changed_is_identical_to_last_run(key) is False:
                    log.info(
                        '+ Filtering unwanted map objects out of map of %s', key)

                    # https://docs.osmcode.org/osmium/latest/osmium-tags-filter.html
                    cmd = ['osmium', 'tags-filter', '--remove-tags']
                    cmd.append(val['map_file'])
                    cmd.extend(translate_tags_to_keep(
                        sys_platform=platform.system()))
                    cmd.extend(['-o', out_file_pbf_filtered_mac])
                    cmd.append('--overwrite')

                    run_subprocess_and_log_output(
                        cmd, f'! Error in Osmium with country: {key}')

                    cmd = ['osmium', 'tags-filter', '--remove-tags']
                    cmd.append(val['map_file'])
                    cmd.extend(translate_tags_to_keep(
                        name_tags=True, sys_platform=platform.system()))
                    cmd.extend(['-o', out_file_pbf_filtered_names_mac])
                    cmd.append('--overwrite')

                    run_subprocess_and_log_output(
                        cmd, f'! Error in Osmium with country: {key}')

                val['filtered_file'] = out_file_pbf_filtered_mac
                val['filtered_file_names'] = out_file_pbf_filtered_names_mac

            # write config file for country
            self.write_country_config_file(key)

        log.info('+ Filter tags from country osm.pbf files: OK')

    def generate_land(self):
        """
        Generate land for all tiles
        """

        log.info('-' * 80)
        log.info('# Generate land for each coordinate')

        tile_count = 1
        for tile in self.o_osm_data.tiles:
            land_file = os.path.join(USER_OUTPUT_DIR,
                                     f'{tile["x"]}', f'{tile["y"]}', 'land.shp')
            out_file_land1 = os.path.join(USER_OUTPUT_DIR,
                                          f'{tile["x"]}', f'{tile["y"]}', 'land')

            # create land.dbf, land.prj, land.shp, land.shx
            if not os.path.isfile(land_file) or self.o_osm_data.force_processing is True:
                log.info(
                    '+ Coordinates: %s,%s. (%s of %s)', tile["x"], tile["y"], tile_count, len(self.o_osm_data.tiles))
                cmd = ['ogr2ogr', '-overwrite', '-skipfailures']
                # Try to prevent getting outside of the +/-180 and +/- 90 degrees borders. Normally the +/- 0.1 are there to prevent white lines at border borders.
                if tile["x"] == 255 or tile["y"] == 255 or tile["x"] == 0 or tile["y"] == 0:
                    cmd.extend(['-spat', f'{tile["left"]:.6f}',
                                f'{tile["bottom"]:.6f}',
                                f'{tile["right"]:.6f}',
                                f'{tile["top"]:.6f}'])
                else:
                    cmd.extend(['-spat', f'{tile["left"]-0.1:.6f}',
                                f'{tile["bottom"]-0.1:.6f}',
                                f'{tile["right"]+0.1:.6f}',
                                f'{tile["top"]+0.1:.6f}'])
                cmd.append(land_file)
                cmd.append(LAND_POLYGONS_PATH)

                run_subprocess_and_log_output(
                    cmd, f'! Error generating land for tile: {tile["x"]},{tile["y"]}')

            # create land1.osm
            if not os.path.isfile(out_file_land1+'1.osm') or self.o_osm_data.force_processing is True:
                # Windows
                if platform.system() == "Windows":
                    cmd = ['python', os.path.join(RESOURCES_DIR,
                                                  'shape2osm.py'), '-l', out_file_land1, land_file]

                # Non-Windows
                else:
                    cmd = ['python', os.path.join(RESOURCES_DIR,
                                                  'shape2osm.py'), '-l', out_file_land1, land_file]

                run_subprocess_and_log_output(
                    cmd, f'! Error creating land.osm for tile: {tile["x"]},{tile["y"]}')
            tile_count += 1

        log.info('+ Generate land for each coordinate: OK')

    def generate_sea(self):
        """
        Generate sea for all tiles
        """

        log.info('-' * 80)
        log.info('# Generate sea for each coordinate')

        tile_count = 1
        for tile in self.o_osm_data.tiles:
            out_file_sea = os.path.join(USER_OUTPUT_DIR,
                                        f'{tile["x"]}', f'{tile["y"]}', 'sea.osm')
            if not os.path.isfile(out_file_sea) or self.o_osm_data.force_processing is True:
                log.info(
                    '+ Coordinates: %s,%s. (%s of %s)', tile["x"], tile["y"], tile_count, len(self.o_osm_data.tiles))
                with open(os.path.join(RESOURCES_DIR, 'sea.osm'), encoding="utf-8") as sea_file:
                    sea_data = sea_file.read()

                    # Try to prevent getting outside of the +/-180 and +/- 90 degrees borders. Normally the +/- 0.1 are there to prevent white lines at tile borders
                    if tile["x"] == 255 or tile["y"] == 255 or tile["x"] == 0 or tile["y"] == 0:
                        sea_data = sea_data.replace(
                            '$LEFT', f'{tile["left"]:.6f}')
                        sea_data = sea_data.replace(
                            '$BOTTOM', f'{tile["bottom"]:.6f}')
                        sea_data = sea_data.replace(
                            '$RIGHT', f'{tile["right"]:.6f}')
                        sea_data = sea_data.replace(
                            '$TOP', f'{tile["top"]:.6f}')
                    else:
                        sea_data = sea_data.replace(
                            '$LEFT', f'{tile["left"]-0.1:.6f}')
                        sea_data = sea_data.replace(
                            '$BOTTOM', f'{tile["bottom"]-0.1:.6f}')
                        sea_data = sea_data.replace(
                            '$RIGHT', f'{tile["right"]+0.1:.6f}')
                        sea_data = sea_data.replace(
                            '$TOP', f'{tile["top"]+0.1:.6f}')

                    with open(out_file_sea, mode='w', encoding="utf-8") as output_file:
                        output_file.write(sea_data)
            tile_count += 1

        log.info('+ Generate sea for each coordinate: OK')

    def split_filtered_country_files_to_tiles(self):
        """
        Split filtered country files to tiles
        """

        log.info('-' * 80)
        log.info('# Split filtered country files to tiles')
        tile_count = 1
        for tile in self.o_osm_data.tiles:

            for country, val in self.o_osm_data.border_countries.items():
                if country not in tile['countries']:
                    continue
                log.info(
                    '+ Coordinates: %s,%s / %s (%s of %s)', tile["x"], tile["y"], country, tile_count, len(self.o_osm_data.tiles))
                out_file = os.path.join(USER_OUTPUT_DIR,
                                        f'{tile["x"]}', f'{tile["y"]}', f'split-{country}.osm.pbf')
                out_file_names = os.path.join(USER_OUTPUT_DIR,
                                              f'{tile["x"]}', f'{tile["y"]}', f'split-{country}-names.osm.pbf')

                # split filtered country files to tiles every time because the result is different per constants (user input)
                # Windows
                if platform.system() == "Windows":
                    cmd = [self.osmconvert_path,
                           '-v', '--hash-memory=2500']
                    cmd.append('-b='+f'{tile["left"]}' + ',' + f'{tile["bottom"]}' +
                               ',' + f'{tile["right"]}' + ',' + f'{tile["top"]}')
                    cmd.extend(
                        ['--complete-ways', '--complete-multipolygons', '--complete-boundaries'])
                    cmd.append(val['filtered_file'])
                    cmd.append('-o='+out_file)

                    run_subprocess_and_log_output(
                        cmd, f'! Error in Osmosis with country: {country}. Win/out_file')

                    cmd = [self.osmconvert_path,
                           '-v', '--hash-memory=2500']
                    cmd.append('-b='+f'{tile["left"]}' + ',' + f'{tile["bottom"]}' +
                               ',' + f'{tile["right"]}' + ',' + f'{tile["top"]}')
                    cmd.extend(
                        ['--complete-ways', '--complete-multipolygons', '--complete-boundaries'])
                    cmd.append(val['filtered_file_names'])
                    cmd.append('-o='+out_file_names)

                    run_subprocess_and_log_output(
                        cmd, '! Error in Osmosis with country: {country}. Win/out_file_names')

                # Non-Windows
                else:
                    cmd = ['osmium', 'extract']
                    cmd.extend(
                        ['-b', f'{tile["left"]},{tile["bottom"]},{tile["right"]},{tile["top"]}'])
                    cmd.append(val['filtered_file'])
                    cmd.extend(['-s', 'smart'])
                    cmd.extend(['-o', out_file])
                    cmd.extend(['--overwrite'])

                    run_subprocess_and_log_output(
                        cmd, '! Error in Osmosis with country: {country}. macOS/out_file')

                    cmd = ['osmium', 'extract']
                    cmd.extend(
                        ['-b', f'{tile["left"]},{tile["bottom"]},{tile["right"]},{tile["top"]}'])
                    cmd.append(val['filtered_file_names'])
                    cmd.extend(['-s', 'smart'])
                    cmd.extend(['-o', out_file_names])
                    cmd.extend(['--overwrite'])

                    run_subprocess_and_log_output(
                        cmd, '! Error in Osmosis with country: {country}. macOS/out_file_names')

            tile_count += 1

        log.info('+ Split filtered country files to tiles: OK')

    def merge_splitted_tiles_with_land_and_sea(self, process_border_countries):
        """
        Merge splitted tiles with land an sea
        """

        log.info('-' * 80)
        log.info('# Merge splitted tiles with land an sea')
        tile_count = 1
        for tile in self.o_osm_data.tiles:  # pylint: disable=too-many-nested-blocks
            log.info(
                '+ Coordinates: %s,%s (%s of %s)', tile["x"], tile["y"], tile_count, len(self.o_osm_data.tiles))

            out_tile_dir = os.path.join(USER_OUTPUT_DIR,
                                        f'{tile["x"]}', f'{tile["y"]}')
            out_file_merged = os.path.join(out_tile_dir, 'merged.osm.pbf')

            land_files = glob.glob(os.path.join(out_tile_dir, 'land*.osm'))

            # merge splitted tiles with land and sea every time because the result is different per constants (user input)
            # sort land* osm files
            self.sort_osm_files(tile)

            # Windows
            if platform.system() == "Windows":
                cmd = [OSMOSIS_WIN_FILE_PATH]
            # Non-Windows
            else:
                cmd = ['osmosis']

            loop = 0
            # loop through all countries of tile, if border-countries should be processed.
            # if border-countries should not be processed, only process the "entered" country
            for country in tile['countries']:
                if process_border_countries or country in self.o_osm_data.border_countries:
                    cmd.append('--rbf')
                    cmd.append(os.path.join(
                        out_tile_dir, f'split-{country}.osm.pbf'))
                    cmd.append('workers=' + self.workers)
                    if loop > 0:
                        cmd.append('--merge')

                    cmd.append('--rbf')
                    cmd.append(os.path.join(
                        out_tile_dir, f'split-{country}-names.osm.pbf'))
                    cmd.append('workers=' + self.workers)
                    cmd.append('--merge')

                    loop += 1

            for land in land_files:
                cmd.extend(
                    ['--rx', 'file='+land, '--s', '--m'])

            cmd.extend(
                ['--rx', 'file='+os.path.join(out_tile_dir, 'sea.osm'), '--s', '--m'])
            cmd.extend(['--tag-transform', 'file=' + os.path.join(RESOURCES_DIR,
                                                                  'tunnel-transform.xml'), '--wb', out_file_merged, 'omitmetadata=true'])

            run_subprocess_and_log_output(
                cmd, f'! Error in Osmosis with tile: {tile["x"]},{tile["y"]}')

            tile_count += 1

        log.info('+ Merge splitted tiles with land an sea: OK')

    def sort_osm_files(self, tile):
        """
        sort land*.osm files to be in this order: nodes, then ways, then relations.
        this is mandatory for osmium-merge since:
        https://github.com/osmcode/osmium-tool/releases/tag/v1.13.2
        """

        log.debug('-' * 80)
        log.debug('# Sorting land* osm files')

        # get all land* osm files
        land_files = glob.glob(os.path.join(USER_OUTPUT_DIR,
                                            f'{tile["x"]}', f'{tile["y"]}', 'land*.osm'))

        for land in land_files:
            if platform.system() == "Windows":
                cmd = [OSMOSIS_WIN_FILE_PATH]
            else:
                cmd = ['osmosis']

            cmd.extend(['--read-xml', 'file='+land])
            cmd.append('--sort')
            cmd.extend(['--write-xml', 'file='+land])

        run_subprocess_and_log_output(
            cmd, f'Error in Osmosis with sorting land* osm files of tile: {tile["x"]},{tile["y"]}')

        log.debug('+ Sorting land* osm files: OK')

    def create_map_files(self, save_cruiser, tag_wahoo_xml):
        """
        Creating .map files
        """

        log.info('-' * 80)
        log.info('# Creating .map files for tiles')

        # Number of threads to use in the mapwriter plug-in
        threads = multiprocessing.cpu_count() - 1
        if int(threads) < 1:
            threads = 1

        tile_count = 1
        for tile in self.o_osm_data.tiles:
            log.info(
                '+ Coordinates: %s,%s (%s of %s)', tile["x"], tile["y"], tile_count, len(self.o_osm_data.tiles))

            out_file_map = os.path.join(USER_OUTPUT_DIR,
                                        f'{tile["x"]}', f'{tile["y"]}.map')

            # apply tag-wahoo xml every time because the result is different per .xml file (user input)
            merged_file = os.path.join(USER_OUTPUT_DIR,
                                       f'{tile["x"]}', f'{tile["y"]}', 'merged.osm.pbf')

            # Windows
            if platform.system() == "Windows":
                cmd = [OSMOSIS_WIN_FILE_PATH, '--rbf', merged_file,
                       'workers=' + self.workers, '--mw', 'file='+out_file_map]
            # Non-Windows
            else:
                cmd = ['osmosis', '--rb', merged_file,
                       '--mw', 'file='+out_file_map]

            cmd.append(
                f'bbox={tile["bottom"]:.6f},{tile["left"]:.6f},{tile["top"]:.6f},{tile["right"]:.6f}')
            cmd.append('zoom-interval-conf=10,0,17')
            cmd.append(f'threads={threads}')
            # add path to tag-wahoo xml file
            try:
                cmd.append(
                    f'tag-conf-file={get_tag_wahoo_xml_path(tag_wahoo_xml)}')
            except TagWahooXmlNotFoundError:
                log.error(
                    'The tag-wahoo xml file was not found: ˚%s˚. Does the file exist and is your input correct?', tag_wahoo_xml)
                sys.exit()

            run_subprocess_and_log_output(
                cmd, f'Error in creating map file via Osmosis with tile: {tile["x"]},{tile["y"]}. mapwriter plugin installed?')

            # Windows
            if platform.system() == "Windows":
                cmd = [get_tooling_win_path('lzma'), 'e', out_file_map,
                       out_file_map+'.lzma', f'-mt{threads}', '-d27', '-fb273', '-eos']
            # Non-Windows
            else:
                # force overwrite of output file and (de)compress links
                cmd = ['lzma', out_file_map, '-f']

                # --keep: do not delete source file
                if save_cruiser:
                    cmd.append('--keep')

            run_subprocess_and_log_output(
                cmd, f'! Error creating map files for tile: {tile["x"]},{tile["y"]}')

            # Create "tile present" file
            with open(out_file_map + '.lzma.17', mode='wb') as tile_present_file:
                tile_present_file.close()

            tile_count += 1

        log.info('+ Creating .map files for tiles: OK')

    def make_and_zip_files(self, extension, zip_folder):
        """
        make or make and zip .map or .map.lzma files
        extension: '.map.lzma' for Wahoo tiles
        extension: '.map' for Cruiser map files
        """

        if extension == '.map.lzma':
            folder_name = self.o_osm_data.country_name
        else:
            folder_name = self.o_osm_data.country_name + '-maps'

        log.info('-' * 80)
        log.info('# Create: %s files', extension)
        log.info('+ Country: %s', self.o_osm_data.country_name)

        # Check for us/utah etc names
        try:
            res = self.o_osm_data.country_name.index('/')
            self.o_osm_data.country_name = self.o_osm_data.country_name[res+1:]
        except ValueError:
            pass

        # copy the needed tiles to the country folder
        log.info('+ Copying %s tiles to output folders', extension)
        for tile in self.o_osm_data.tiles:
            src = os.path.join(f'{USER_OUTPUT_DIR}',
                               f'{tile["x"]}', f'{tile["y"]}') + extension
            dst = os.path.join(
                f'{USER_WAHOO_MC}', folder_name, f'{tile["x"]}', f'{tile["y"]}') + extension
            self.copy_to_dst(extension, src, dst)

            if extension == '.map.lzma':
                src = src + '.17'
                dst = dst + '.17'
                self.copy_to_dst(extension, src, dst)

        if zip_folder:
            # Windows
            if platform.system() == "Windows":
                cmd = [get_tooling_win_path('7za'), 'a', '-tzip']

                cmd.extend(
                    [folder_name + '.zip', os.path.join(".", folder_name, "*")])

            # Non-Windows
            else:
                cmd = ['zip', '-r']

                cmd.extend(
                    [folder_name + '.zip', folder_name])

            run_subprocess_and_log_output(
                cmd, f'! Error zipping map files for folder: {folder_name}', cwd=USER_WAHOO_MC)

            # Delete the country/region map folders after compression
            try:
                shutil.rmtree(os.path.join(
                    f'{USER_WAHOO_MC}', folder_name))
            except OSError:
                log.error(
                    '! Error, could not delete folder %s', os.path.join(USER_WAHOO_MC, folder_name))

            log.info('+ Zip %s files: OK', extension)

        log.info('+ Create %s files: OK', extension)

    def copy_to_dst(self, extension, src, dst):
        """
        Zip .map or .map.lzma files
        postfix: '.map.lzma' for Wahoo tiles
        postfix: '.map' for Cruiser map files
        """
        outdir = os.path.dirname(dst)

        # first create the to-directory if not already there
        os.makedirs(outdir, exist_ok=True)

        try:
            shutil.copy2(src, dst)
        except Exception as exception:  # pylint: disable=broad-except
            log.error(
                '! Error copying %s files for country %s: %s', extension, self.o_osm_data.country_name, exception)
            sys.exit()

    def write_country_config_file(self, country):
        """
        Write country config file into _tiles/{country} directory
        """
        # Data to be written
        configuration = {
            "version_last_run": VERSION,
            "changed_ts_map_last_run": get_timestamp_last_changed(self.o_osm_data.border_countries[country]['map_file']),
            "tags_last_run": translate_tags_to_keep(sys_platform=platform.system()),
            "name_tags_last_run": translate_tags_to_keep(name_tags=True, sys_platform=platform.system())
        }

        write_json_file_generic(os.path.join(
            USER_OUTPUT_DIR, country, ".config.json"), configuration)

    def tags_are_identical_to_last_run(self, country):
        """
        compare tags of this run with used tags from last run stored in _tiles/{country} directory
        """
        tags_are_identical = True

        try:
            country_config = read_json_file(os.path.join(
                USER_OUTPUT_DIR, country, ".config.json"))
            if not country_config["tags_last_run"] == translate_tags_to_keep(sys_platform=platform.system()) \
                    or not country_config["name_tags_last_run"] == translate_tags_to_keep(name_tags=True, sys_platform=platform.system()):
                tags_are_identical = False
        except (FileNotFoundError, KeyError):
            tags_are_identical = False

        return tags_are_identical

    def last_changed_is_identical_to_last_run(self, country):
        """
        compare tags of this run with used tags from last run stored in _tiles/{country} directory
        """
        last_changed_is_identical = True

        try:
            country_config = read_json_file(os.path.join(
                USER_OUTPUT_DIR, country, ".config.json"))
            if not country_config["changed_ts_map_last_run"] == get_timestamp_last_changed(self.o_osm_data.border_countries[country]['map_file']):
                last_changed_is_identical = False
        except (FileNotFoundError, KeyError):
            last_changed_is_identical = False

        return last_changed_is_identical
